"""
    This module is a Indexed Berkeley DB.
    
    @author: Sury Prakash Soni (suryasoni@gmail.com)

"""
import time, simplejson, os, random, re, string, sys, threading, optparse, traceback
from bsddb3.db  import *
from indexer.generic import *
from wrappers import CursorWrapper

class DBTable():
    """
        BDB Indexing wrapper class.
    """
    def __init__(self, dbname, indexSetup):
        self.envFlags           = DB_CREATE|DB_INIT_LOCK|DB_INIT_LOG|DB_INIT_MPOOL|DB_INIT_TXN|DB_THREAD
        self.envFlagsStr        = "DB_CREATE|DB_INIT_LOCK|DB_INIT_LOG|DB_INIT_MPOOL|DB_INIT_TXN|DB_THREAD"
        self.dbname             = dbname
        self.indexSetup         = indexSetup
        self.env                = None
        self.db                 = None
        self.dbs                = []
        self.counter            = 1
        self.dbhome             = "dbs" 

        self.counterLock        = threading.Condition()

        self.env = DBEnv()
        """
            To get performance advantage, we can disable SYNC'ing feature.
        """
        self.env.set_flags(DB_TXN_NOSYNC, 1)

        """
            Make sure, database directory available.
        """
        if not os.path.isdir(self.dbhome):
            os.mkdir(self.dbhome)
        
        home = os.path.join(self.dbhome, self.dbname)

        if not os.path.isdir(home):
            os.mkdir(home)
        
        self.env.open(home, self.envFlags, 0)

        #self.env.set_cachesize(0, 10*1024*1024, 0)
        mydbtype    = DB_BTREE
        dbFlags     = DB_CREATE | DB_AUTO_COMMIT
        self.columnIndexer = ColumnIndexer(self.indexSetup)

        try:
            self.db = DB(dbEnv=self.env)
            self.db.set_bt_compare(self.columnIndexer.customCmp)
            self.db.open("master.db", dbtype=mydbtype, flags=dbFlags)

            """
                Initialise secondary database handles.
            """
            self.dbs = self.columnIndexer.getSecondaryDBS(self.env, self.db)
        except:
            raise
        
    def cleanup(self, recover=0):
        self.close()

        try:
            os.rmdir(self.dbhome)
        except:
            pass

    def close(self):
        """
            Close database artifacts in following order:
            Close all secondary database handles.
            Close primary database handle.
            close environment handle.
        """
        try:
            for (columnName, dbsi) in self.dbs.iteritems():
                dbsi.close()
        except:
            raise

        try:
            self.db.close()
        except:
            raise
        try:
            self.env.close()
        except:
            raise

    def append(self, data):
        self.counterLock.acquire()
        try:
            key = str(self.counter)
            self.counter += 1
        finally:
            self.counterLock.release()
            
        txnobj = self.env.txn_begin(flags=DB_TXN_NOSYNC)
        try:
            #print "key:",key,",data:",data
            self.db.put(key, simplejson.dumps(data), txn=txnobj)
            txnobj.commit()
        except:
            print "exception in put! key:",key
            txnobj.abort()
            raise
        
    def delete(self, key):
        """
            Function to delete key,value pair

            TODO: check for commits policy
            
            @type key: string
            @param key: key
            
            @note: check for commits policy
        """
        txnobj = self.env.txn_begin(flags=DB_TXN_NOSYNC)
        try:
            deleted = False
            if self.dbflags.exists(key, txn=txnobj):
                (k,v) = self.dbflags.get(key, txn=txnobj)
                if v=="0":
                    self.dbflags.put(key, "1", txn=txnobj)
                    deleted = True
            txnobj.commit()
            return deleted
        except:
            txnobj.abort()
            raise NextDigital.DBReadOnlyException()

    def search(self, pattern=None):
        """
            This interface is bit tricky.
            Here we investigate partial keys as primary
            key for secondary databases.
            We join all secondary database cursors
            and return combined (natural join) cursor.
    
            TODO: If we need to support different version
            of key formats, then this is the place, which
            need to be very carefully taken care of.
            
            @type pattern: string
            @param pattern: search pattern
        """

        try:
            if pattern is None:
                """
                    If no secondary databases available,
                    return cursor on primary database.
                    TODO: potentially initial position on primary
                    cursor could be set for partial search patterns.
                """
                txnobj = self.env.txn_begin(flags=DB_TXN_NOSYNC)
                cur    = self.db.cursor(txn=txnobj)
                return CursorWrapper(txnobj, [], cur)

            elif len(self.dbs)>0:
                queries = self.columnIndexer.parse(pattern)
                
                curs        = []
                i=0
                txnobj = self.env.txn_begin(flags=DB_TXN_NOSYNC)

                """
                    Get all individual secondary database cursors
                    all set to first found resulting key.
                """
                
                for (columnName, (condOp, value)) in pattern.iteritems():
                    try:
                        cur = self.dbs[columnName].cursor(txn=txnobj)
                        if value:
                            if condOp=="eq":
                                cur.set(key=str(value))
                            elif condOp=="like":
                                cur.set_range(key=str(value))
                        else:
                            cur.first()
                        try:
                            if cur.current():
                                #print "current1:",cur.current()
                                curs.append(cur)
                            else:
                                cur.close()
                        except DBInvalidArgError, e:
                            cur.close()
                            for cur in curs:
                                cur.close()
                            traceback.print_exc(file=sys.stdout)
                            raise e
                    except:
                        traceback.print_exc(file=sys.stdout)
                        raise

                """
                    If none of the secondary database contains keys
                    matching search pattern, then obviously primary
                    database do not contain requested items. Hence,
                    raise the exception.
                """
                if len(curs)<=0:
                    raise

                """
                    If, few of the secondary cursors match the pattern.
                    Join them on primary database and return joinedCursor.
                """
                try:
                    joinedCursor    = self.db.join(curs)
                    #print "main joinedCursor data:", joinedCursor.get(0)
                    #print "joinedCursor:", joinedCursor
                    if not joinedCursor:
                        raise
                    return CursorWrapper(txnobj, curs, joinedCursor)
                except Exception, e:
                    traceback.print_exc(file=sys.stdout)
                    raise
        except Exception, e:
            traceback.print_exc(file=sys.stdout)
            raise


    def records(self):
        """
            Function to get number of items in the DB.
            
            @rtype: string
            @return: number of items in the DB.
        """
        try:
            return str(self.db.stat()['ndata'])
        except:
            return "0"
