"""
    Parser module consist of BDB custom indexing classes.

    @author: Sury Prakash Soni (suryasoni@gmail.com)
"""
from bsddb3.db import *
from string import strip
import simplejson, sys, traceback

class BaseSKIndexer:
    """
        This is the base class for the secondary key indexer.
    """
    def getSKey(self, key, value):
        """
            This function returns the secondary key for given master, key, value.
            @type key: string
            @param key: Primary DB key
            
            @type value: string
            @param value: primary key's values
            
            @rtype: string
            @return: secondary DB key
        """
        raise NotImplementedError()

class BaseIndexer:
    """
        This is the base class for primary DB key indexer.
    """
    def customCmp(self, x, y):
        """
            This function will compare the keys of two primary DB items.
            @type x: string
            @param x: Application/user passed key
            
            @type y: string
            @param y: current item in the given DB.
            
            @rtype: int
            @return: (-1, 0, 1)
        """
        raise NotImplementedError()
    
    def parse(self, key):
        """
            Function to parse the key
            
            @type key: string
            @param key: key to be parsed
            
            @rtype: list
            @return: list of parsed key parts
        """
        raise NotImplementedError()
    
    def verify(self, key):
        """
            Function to check whether given key is valid.
            
            @type key: string
            @param key: key to verify
            
            @rtype: boolean
            @return: boolean whether key is valid or not!
        """
        raise NotImplementedError()

    def isAny(self, key):
        """
            Function to check whether key is a valid any string
            
            @type key: string
            @param key: data to match against any string
            
            @rtype: boolean
            @return: result whether it is a valid any string.
        """
        raise NotImplementedError()
    
    def getSecondary(self, index):
        """
            Function to get secondary index key extractor object.
            
            @type index: int
            @param index: index level for which extractor object need
            to be returned.
            
            @rtype: function
            @return: secondary key extractor object funciton.
        """
        raise NotImplementedError()

class StringSKIndexer(BaseSKIndexer):
    """
        String based secondary key indexer - BaseSKIndexer
        interface implementation class
    """
    def __init__(self, skIndexLevel, seperator="$"):
        if skIndexLevel<1:
            raise KeyError()
        self.skIndexLevel = skIndexLevel
        self.seperator = seperator
        
    def getSKey(self, key, value):
        parts = key.split(self.seperator)
        if len(parts)>= self.skIndexLevel:
            return parts[self.skIndexLevel-1]

class StringIndexer(BaseIndexer):
    """
        String based indexer interface.
    """
    def __init__(self, indexLevel, seperator="$", any="*"):
        """
            Constructor
            
            @type indexLevel: int
            @param indexLevel: number of indices on the database
            
            @type seperator: string
            @param seperator: key format field seperator string
            
            @type any: string
            @param any: any string representor.
        """
        self.indexLevel = int(indexLevel)
        self.seperator = seperator
        self.any = any
        self.debug = False

    def customCmp(self, x, y):
        """
            This function will compare the keys of two primary DB items.
            @type x: string
            @param x: Application/user passed key
            
            @type y: string
            @param y: current item in the given DB.
            
            @rtype: int
            @return: (-1, 0, 1)
        """
        if self.debug:
            print "Comparing: x: ", x,", y: ", y

        if self.indexLevel==0:
            if x=="" and y=="":
                return 0
            return int(x) - int(y)
        
        partsX = self.parse(x)
        partsY = self.parse(y)
        
        steps = len(partsX)
        result = 0
        for i in xrange(0, steps):
            if partsX[i]==self.any or partsX[i] == partsY[i]:
                result = 0
                continue
            elif partsX[i] < partsY[i]:
                result = -1
                break
            elif partsX[i] > partsY[i]:
                result = 1
                break
            
        return result

    def parse(self, key):
        """
            Function to parse the key
            
            @type key: string
            @param key: key to be parsed
            
            @rtype: list
            @return: list of parsed key parts
        """
        try:
            return map(strip, key.split(self.seperator))
        except:
            return [key]
    
    def join(self, parts):
        return self.seperator.join(parts)
    
    def verify(self, key):
        """
            Function to check whether given key is valid.
            
            @type key: string
            @param key: key to verify
            
            @rtype: boolean
            @return: boolean whether key is valid or not!
        """
        parts = self.parse(key)
        return len(parts) == self.indexLevel

    def isAny(self, key):
        """
            Function to check whether key is a valid any string
            
            @type key: string
            @param key: data to match against any string
            
            @rtype: boolean
            @return: result whether it is a valid any string.
        """
        return key==self.any
    
    def getSecondary(self, index):
        """
            Function to get secondary index key extractor object.
            
            @type index: int
            @param index: index level for which extractor object need
            to be returned.
            
            @rtype: function
            @return: secondary key extractor object funciton.
        """
        skIndexer = StringSKIndexer(index, self.seperator)
        return skIndexer.getSKey

class ColumnIndexer(BaseIndexer):
    def __init__(self, indexSetup, indexSetupEncoding="", valueEncoding="json"):
        self.indexSetup            = indexSetup
        self.indexSetupEncoding    = unicode(indexSetupEncoding).lower()
        self.valueEncoding          = unicode(valueEncoding).lower()
        if self.indexSetupEncoding=="json":
            self.indexSetup = simplejson.loads(indexSetup)
    
    def getSecondaryDBS(self, env, db):
        dbFlags     = DB_CREATE | DB_AUTO_COMMIT
        myindextype = DB_BTREE
        dbs = {}
        for (columnName, columnType) in self.indexSetup.iteritems():
            dbs[columnName] = DB(dbEnv=env)
            if columnType=="int":
                dbs[columnName].set_bt_compare(self.intCmp)
            elif columnType=="float":
                dbs[columnName].set_bt_compare(self.floatCmp)
            dbs[columnName].set_flags(DB_DUP|DB_DUPSORT)
            dbs[columnName].open("secondary_"+str(columnName)+".db", dbtype=myindextype, flags=dbFlags)
            db.associate(dbs[columnName], self.getSecondary(columnName).getSKey)
        return dbs
    
    def getSecondary(self, columnName):
        columnType = self.indexSetup[columnName]
        
        return ColumnSKIndexer(columnName, columnType, self.valueEncoding)

    def intCmp(self, x, y):
        if x=="" and y=="":
            return 0
        #print "Comparing",x,"with",y

        pattern=False
        if x.endswith("a"):
            pattern="a"
        elif x.endswith("d"):
            pattern="d"
        
        x = x.strip("ad")
        
        ret = 0
        
        if (int(x)-int(y))<0:
            if pattern<>"a":
                ret = -1
            else:
                print "special case"
        elif (int(x)-int(y))>0:
            if pattern<>"d":
                ret = 1
            else:
                print "special case"
        return ret

    def floatCmp(self, x, y):
        if x=="" and y=="":
            return 0
        return float(x)-float(y)

    def customCmp(self, x, y):
        """
            This function will compare the keys of two primary DB items.
            @type x: string
            @param x: Application/user passed key
            
            @type y: string
            @param y: current item in the given DB.
            
            @rtype: int
            @return: (-1, 0, 1)
        """
        if x=="" and y=="":
            return 0
        
        ret = 0

        if(int(x)-int(y)<0):
            ret=-1
        elif(int(x)-int(y)>0):
            ret=1
        return ret

    def parse(self, pattern):
        """
            Function to parse the key
            
            @type key: string
            @param key: key to be parsed
            
            @rtype: list
            @return: list of parsed key parts
        """
        return pattern



class ColumnSKIndexer(BaseSKIndexer):
    """
        Class to handle secondary index key generation from primary key, value pair.
    """
    def __init__(self, columnName, columnType, valueEncoding):
        """
            Constructor.
            
            @type columnName: string
            @param columnName: Name of the column on which indexing need to be done.
            
            @type valueEncoding: string
            @param valueEncoding: encoding type of the primary value
        """
        self.columnName = columnName
        self.columnType = columnType
        self.valueEncoding = unicode(valueEncoding).lower()
    
    def getSKey(self, key, value):
        """
            Function to generate the secondary key.
            
            @type key: string
            @param key: primary key
            
            @type value: string
            @param value: primary value
            
            @rtype: string
            @return: returns secondary key fetched from the primary value.
        """
        try:
            if self.valueEncoding == "json":
                value = simplejson.loads(value)
            skey = value
            try:
                skey = value[self.columnName]
            except:
                raise
            
            skey = str(skey)
            #print "skey:", skey,", columnName:", self.columnName
            return skey
        except:
            traceback.print_exc(file=sys.stdout)
            return DB_DONOTINDEX
